@@ -14,27 +14,27 @@
 """v3iofs - An fsspec driver for v3io"""
 
 __all__ = [
-    '__version__',
-    'V3ioFS',
-    'V3ioFile',
+    "__version__",
+    "V3ioFS",
+    "V3ioFile",
 ]
 
-__version__ = '0.1.13'
+__version__ = '0.1.15'
 
 import fsspec
 
 from .file import V3ioFile  # noqa: F401
 from .fs import V3ioFS  # noqa: F401
 
-if hasattr(fsspec, 'register_implementation'):
+if hasattr(fsspec, "register_implementation"):
     # TODO: Not sure about clobber=True
-    fsspec.register_implementation('v3io', V3ioFS, clobber=True)
+    fsspec.register_implementation("v3io", V3ioFS, clobber=True)
 else:
     from fsspec.registry import known_implementations
 
-    known_implementations['v3io'] = {
-        'class': 'v3iofs.V3ioFS',
-        'err': 'Please install v3iofs to use the v3io fileysstem class'
+    known_implementations["v3io"] = {
+        "class": "v3iofs.V3ioFS",
+        "err": "Please install v3iofs to use the v3io fileysstem class",
     }
 
     del known_implementations
